{% extends 'wildewidgets/block.html' %}
{% load wildewidgets %}

{% block block_content %}
  {% if header or header_text %}
    <div class="card-header {{header_css}}">
      {% if header %}
        {% wildewidgets header %}
      {% elif header_text %}
        {{ header_text}}
      {% endif %}
    </div>
  {% endif %}
  <div class="card-body {{body_css}}">
    {% if title %}
      <div class="card-title">{{ title }}</div>
      {% if subtitle %}
        <div class="card-subtitle">{{ subtitle }}</div>
      {% endif %}
    {% endif %}
    <div class="{{widget_css}}">
      {% wildewidgets widget %}
    </div>
  </div>
{% endblock %}