<svg width="{{fullsize}}" height="{{fullsize}}">
  <circle cx="{{halfsize}}" cy="{{halfsize}}" r="{{halfsize}}" fill="{{background_color}}"><title>{{fullname}}</title></circle>
  <text x="50%" y="50%" text-anchor="middle" fill="{{color}}" font-size="{{halfsize}}px" font-family="Arial" dy=".3em"><title>{{fullname}}</title>{{initials}}</text>
</svg>